"""Module containing the definition of colors used in the app."""
from typing import Final

from colorama import Fore

yt_col: Final[str] = Fore.MAGENTA
md_col: Final[str] = Fore.CYAN
man_col: Final[str] = Fore.GREEN
auto_col: Final[str] = Fore.YELLOW
